@@ -49,9 +49,9 @@ typedef struct {
 
 static PVector* EMPTY_VECTOR = NULL;
 static PyObject* pmap_factory_fn = NULL;
-static PyObject* assoc_in_fn_name = NULL;
+static PyObject* set_in_fn_name = NULL;
 
-static PyObject* assocInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject* value) {
+static PyObject* setInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject* value) {
   if(pmap_factory_fn == NULL) {
     // Lazy import pmap factory to avoid circular import problems
     pmap_factory_fn = PyObject_GetAttrString(PyImport_ImportModule("pyrsistent"), "pmap");
@@ -60,7 +60,7 @@ static PyObject* assocInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject
   PyObject* emptyMap = PyObject_CallFunctionObjArgs(pmap_factory_fn, NULL);
   PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
   PyObject* newMap = PyObject_CallMethodObjArgs(emptyMap,
-						assoc_in_fn_name,
+						set_in_fn_name,
 						newSequence,
 						value,
 						NULL);
@@ -226,27 +226,34 @@ static void PVector_dealloc(PVector *self) {
 }
 
 static PyObject *PVector_repr(PVector *self) {
-  // Reuse the tuple repr code, a bit less efficient but saves some code
+  // Reuse the list repr code, a bit less efficient but saves some code
   Py_ssize_t i;
-  i = Py_ReprEnter((PyObject *)self);
-  if (i != 0) {
-    return i > 0 ? PyString_FromString("(...)") : NULL;
-  }
-
-  PyObject *tuple = PyTuple_New(self->count);
+  PyObject *list = PyList_New(self->count);
   for (i = 0; i < self->count; ++i) {
     PyObject *o = _get_item(self, i);
     Py_INCREF(o);
-    PyTuple_SET_ITEM(tuple, i, o);
+    PyList_SET_ITEM(list, i, o);
   }
 
-  PyObject *result = PyObject_Repr(tuple);
 
-  Py_DECREF(tuple);
-  Py_ReprLeave((PyObject *)self);
-  return result;
+  PyObject *list_repr = PyObject_Repr(list);
+  Py_DECREF(list);
+
+  // Repr for list implemented differently in python 2 and 3. Need to
+  // handle this or core dump will occur.
+#if PY_MAJOR_VERSION >= 3
+  PyObject *s = PyUnicode_FromFormat("%s%U%s", "pvector(", list_repr, ")");
+  Py_DECREF(list_repr);
+#else
+  PyObject *s = PyString_FromString("pvector(");
+  PyString_ConcatAndDel(&s, list_repr);
+  PyString_ConcatAndDel(&s, PyString_FromString(")"));
+#endif
+
+  return s;
 }
 
+
 static long PVector_hash(PVector *self) {
   // Follows the pattern of the tuple hash
   long x, y;
@@ -413,7 +420,11 @@ static PyObject* PVector_index(PVector *self, PyObject *args) {
   for (i = start; i < stop && i < self->count; i++) {
     int cmp = PyObject_RichCompareBool(_get_item(self, i), value, Py_EQ);
     if (cmp > 0) {
+#if PY_MAJOR_VERSION >= 3
+      return PyLong_FromSsize_t(i);
+#else
       return PyInt_FromSsize_t(i);
+#endif
     } else if (cmp < 0) {
       return NULL;
     }
@@ -436,7 +447,11 @@ static PyObject* PVector_count(PVector *self, PyObject *value) {
     }
   }
 
-  return PyInt_FromSsize_t(count);
+#if PY_MAJOR_VERSION >= 3
+      return PyLong_FromSsize_t(count);
+#else
+      return PyInt_FromSsize_t(count);
+#endif
 }
 
 static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
@@ -446,9 +461,9 @@ static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
 
 static PyObject* PVector_append(PVector *self, PyObject *obj);
 
-static PyObject* PVector_assoc_in(PVector *self, PyObject *obj);
+static PyObject* PVector_set_in(PVector *self, PyObject *obj);
 
-static PyObject* PVector_assoc(PVector *self, PyObject *obj);
+static PyObject* PVector_set(PVector *self, PyObject *obj);
 
 static PyObject* PVector_subscript(PVector* self, PyObject* item);
 
@@ -482,10 +497,10 @@ PyDoc_STRVAR(count_doc,
 
 static PyMethodDef PVector_methods[] = {
 	{"append",      (PyCFunction)PVector_append, METH_O,       "Appends an element"},
-	{"assoc",       (PyCFunction)PVector_assoc, METH_VARARGS, "Inserts an element at the specified position"},
+	{"set",         (PyCFunction)PVector_set, METH_VARARGS, "Inserts an element at the specified position"},
 	{"__getitem__", (PyCFunction)PVector_subscript, METH_O|METH_COEXIST, "Subscript"},
 	{"extend",      (PyCFunction)PVector_extend, METH_O|METH_COEXIST, "Extend"},
-	{"assoc_in",    (PyCFunction)PVector_assoc_in, METH_VARARGS, "Insert an element in a nested structure"},
+	{"set_in",      (PyCFunction)PVector_set_in, METH_VARARGS, "Insert an element in a nested structure"},
 	{"index",       (PyCFunction)PVector_index,  METH_VARARGS, index_doc},
 	{"count",       (PyCFunction)PVector_count,  METH_O, count_doc},
 	{NULL}
@@ -494,9 +509,8 @@ static PyMethodDef PVector_methods[] = {
 static PyObject * PVectorIter_iter(PyObject *seq);
 
 static PyTypeObject PVectorType = {
-  PyObject_HEAD_INIT(NULL)
-  0,
-  "pvectorc.PVector",                       /* tp_name        */
+  PyVarObject_HEAD_INIT(NULL, 0)
+  "pvectorc.PVector",                         /* tp_name        */
   sizeof(PVector),                            /* tp_basicsize   */
   0,		                              /* tp_itemsize    */
   (destructor)PVector_dealloc,                /* tp_dealloc     */
@@ -659,6 +673,13 @@ static void extend_with_item(PVector *newVec, PyObject *item) {
 }
 
 
+#if PY_MAJOR_VERSION >= 3
+// This was changed in 3.2 but we do not claim compatibility with any older version of python 3.
+#define SLICE_CAST
+#else
+#define SLICE_CAST (PySliceObject *)
+#endif
+
 static PyObject *PVector_subscript(PVector* self, PyObject* item) {
     if (PyIndex_Check(item)) {
       Py_ssize_t i = PyNumber_AsSsize_t(item, PyExc_IndexError);
@@ -669,7 +690,7 @@ static PyObject *PVector_subscript(PVector* self, PyObject* item) {
       return PVector_get_item(self, i);
     } else if (PySlice_Check(item)) {
       Py_ssize_t start, stop, step, slicelength, cur, i;
-      if (PySlice_GetIndicesEx((PySliceObject *)item, self->count,
+      if (PySlice_GetIndicesEx(SLICE_CAST item, self->count,
              &start, &stop, &step, &slicelength) < 0) {
          return NULL;
       }
@@ -691,7 +712,7 @@ static PyObject *PVector_subscript(PVector* self, PyObject* item) {
         return (PyObject*)newVec;
       }
     } else {
-      PyErr_Format(PyExc_TypeError, "pvector indices must be integers, not %.200s", item->ob_type->tp_name);
+      PyErr_Format(PyExc_TypeError, "pvector indices must be integers, not %.200s", Py_TYPE(item)->tp_name);
       return NULL;
     }
 } 
@@ -728,7 +749,7 @@ static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
         return NULL;
     }
     
-    iternext = *it->ob_type->tp_iternext;
+    iternext = *Py_TYPE(it)->tp_iternext;
     PyObject *item = iternext(it);
     if (item == NULL) {
       Py_DECREF(it);
@@ -791,26 +812,26 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
   return (PyObject*)pvec;
 }
 
-static VNode* doAssoc(VNode* node, unsigned int level, unsigned int position, PyObject* value) {
+static VNode* doSet(VNode* node, unsigned int level, unsigned int position, PyObject* value) {
   if(level == 0) {
-    debug("doAssoc(): level == 0\n");
+    debug("doSet(): level == 0\n");
     VNode* theNewNode = newNode();
     copyInsert(theNewNode->items, node->items, position & BIT_MASK, value);
     incRefs((PyObject**)theNewNode->items);
     return theNewNode;
   } else {
-    debug("doAssoc(): level == %i\n", level);
+    debug("doSet(): level == %i\n", level);
     VNode* theNewNode = copyNode(node);
     Py_ssize_t index = (position >> level) & BIT_MASK;
 
     // Drop reference to this node since we're about to replace it
     ((VNode*)theNewNode->items[index])->refCount--;
-    theNewNode->items[index] = doAssoc(node->items[index], level - SHIFT, position, value); 
+    theNewNode->items[index] = doSet(node->items[index], level - SHIFT, position, value); 
     return theNewNode;
   }
 }
 
-static PyObject* internalAssoc(PVector *self, Py_ssize_t position, PyObject *argObj) {
+static PyObject* internalSet(PVector *self, Py_ssize_t position, PyObject *argObj) {
   if(position < 0) {
     position += self->count;
   }
@@ -825,7 +846,7 @@ static PyObject* internalAssoc(PVector *self, Py_ssize_t position, PyObject *arg
       return (PyObject*)new_pvec;
     } else {
       // Keep the tail, replace the root
-      VNode *newRoot = doAssoc(self->root, self->shift, position, argObj);
+      VNode *newRoot = doSet(self->root, self->shift, position, argObj);
       PVector *new_pvec = newPvec(self->count, self->shift, newRoot);
 
       // Free the tail and replace it with a reference to the tail of the original vector
@@ -842,7 +863,7 @@ static PyObject* internalAssoc(PVector *self, Py_ssize_t position, PyObject *arg
   }
 }
 
-static PyObject* PVector_assoc_in(PVector *self, PyObject *args) {
+static PyObject* PVector_set_in(PVector *self, PyObject *args) {
   PyObject *keySequence = NULL;
   PyObject *value = NULL;
 
@@ -863,9 +884,9 @@ static PyObject* PVector_assoc_in(PVector *self, PyObject *args) {
     }
 
     if(keySize == 1) {
-      return internalAssoc(self, keyIndex, value);
+      return internalSet(self, keyIndex, value);
     } else if(keyIndex == self->count) {
-      PyObject* newMap = assocInPmap(keySequence, keySize, value);
+      PyObject* newMap = setInPmap(keySequence, keySize, value);
       if(newMap == NULL) {
       	return NULL;
       }
@@ -880,14 +901,14 @@ static PyObject* PVector_assoc_in(PVector *self, PyObject *args) {
       }
 
       PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
-      PyObject* newItem = PyObject_CallMethodObjArgs(currentItem, assoc_in_fn_name, newSequence, value, NULL);
+      PyObject* newItem = PyObject_CallMethodObjArgs(currentItem, set_in_fn_name, newSequence, value, NULL);
       Py_DECREF(currentItem);
       Py_DECREF(newSequence);
       if(newItem == NULL) {
 	return NULL;
       }
       
-      PyObject* newVector = internalAssoc(self, keyIndex, newItem);
+      PyObject* newVector = internalSet(self, keyIndex, newItem);
       Py_DECREF(newItem);
       return newVector;
     }
@@ -897,7 +918,7 @@ static PyObject* PVector_assoc_in(PVector *self, PyObject *args) {
 /*
  Steals a reference to the object that is inserted in the vector.
 */
-static PyObject* PVector_assoc(PVector *self, PyObject *args) {
+static PyObject* PVector_set(PVector *self, PyObject *args) {
   PyObject *argObj = NULL;  /* argument to insert */
   Py_ssize_t position;
 
@@ -906,7 +927,7 @@ static PyObject* PVector_assoc(PVector *self, PyObject *args) {
     return NULL;
   }
 
-  return internalAssoc(self, position, argObj);
+  return internalSet(self, position, argObj);
 }
 
 
@@ -915,7 +936,22 @@ static PyMethodDef PyrsistentMethods[] = {
   {NULL, NULL, 0, NULL}
 };
 
-PyMODINIT_FUNC initpvectorc(void) {
+
+#if PY_MAJOR_VERSION >= 3
+  static struct PyModuleDef moduledef = {
+    PyModuleDef_HEAD_INIT,
+    "pvectorc",          /* m_name */
+    "Persistent vector", /* m_doc */
+    -1,                  /* m_size */
+    PyrsistentMethods,   /* m_methods */
+    NULL,                /* m_reload */
+    NULL,                /* m_traverse */
+    NULL,                /* m_clear */
+    NULL,                /* m_free */
+  };
+#endif
+
+PyObject* moduleinit(void) {
   PyObject* m;
   
   // Only allow creation/initialization through factory method pvec
@@ -923,14 +959,14 @@ PyMODINIT_FUNC initpvectorc(void) {
   PVectorType.tp_new = NULL;
 
   if (PyType_Ready(&PVectorType) < 0)
-    return;
+    return NULL;
 
   // Register with Sequence to appear as a proper sequence to the outside world
   PyObject* c = PyObject_GetAttrString(PyImport_ImportModule("collections"), "Sequence");
 
-  if (PyType_Ready(c->ob_type) < 0) {
+  if (PyType_Ready(Py_TYPE(c)) < 0) {
     printf("Failed to initialize Sequence!\n");
-    return;
+    return NULL;
   }
 
   if(c == NULL) {
@@ -939,9 +975,14 @@ PyMODINIT_FUNC initpvectorc(void) {
   
   PyObject_CallMethod(c, "register", "O", &PVectorType);
 
-  m = Py_InitModule("pvectorc", PyrsistentMethods);
+#if PY_MAJOR_VERSION >= 3
+  m = PyModule_Create(&moduledef);
+#else
+  m = Py_InitModule3("pvectorc", PyrsistentMethods, "Persistent vector");  
+#endif
+
   if (m == NULL) {
-    return;
+    return NULL;
   }
 
   SHIFT = __builtin_popcount(BIT_MASK);
@@ -955,9 +996,22 @@ PyMODINIT_FUNC initpvectorc(void) {
   Py_INCREF(&PVectorType);
   PyModule_AddObject(m, "PVector", (PyObject *)&PVectorType);
 
-  assoc_in_fn_name = PyString_FromString("assoc_in");
+
+  set_in_fn_name = PyUnicode_FromString("set_in");
+
+  return m;
 }
 
+#if PY_MAJOR_VERSION >= 3
+PyMODINIT_FUNC PyInit_pvectorc(void) {
+  return moduleinit();
+}
+#else
+PyMODINIT_FUNC initpvectorc(void) {
+  moduleinit();
+}
+#endif
+
 
 /*********************** PVector Iterator **************************/
 
