@@ -1,7 +1,10 @@
 from collections import Sequence, Mapping, Set, Hashable
 from itertools import chain
-from traceback import extract_stack
 from functools import wraps
+from numbers import Integral
+
+import six
+
 
 def _bitcount(val):
     return bin(val).count("1")
@@ -22,7 +25,7 @@ def _comparator(f):
 
 class PVector(object):
     """
-    Do not instantiate directly, instead use the factory functions :py:func:`pvec` and :py:func:`pvector` to
+    Do not instantiate directly, instead use the factory functions :py:func:`v` and :func:`pvector` to
     create an instance.
 
     Heavily influenced by the persistent vector available in Clojure. Initially this was more or
@@ -37,21 +40,23 @@ class PVector(object):
     semantics are the same (more or less) the same function names have been used but for some cases it is not possible,
     for example assignments.
 
-    The following are examples of some common operations on persistent vectors
+    The PVector implements the Sequence protocol and is Hashable.
+
+    The following are examples of some common operations on persistent vectors:
 
     >>> p = v(1, 2, 3)
     >>> p2 = p.append(4)
     >>> p3 = p2.extend([5, 6, 7])
     >>> p
-    (1, 2, 3)
+    pvector([1, 2, 3])
     >>> p2
-    (1, 2, 3, 4)
+    pvector([1, 2, 3, 4])
     >>> p3
-    (1, 2, 3, 4, 5, 6, 7)
+    pvector([1, 2, 3, 4, 5, 6, 7])
     >>> p3[5]
     6
-    >>> p.assoc(1, 99)
-    (1, 99, 3)
+    >>> p.set(1, 99)
+    pvector([1, 99, 3])
     >>>
     """
     __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset')
@@ -68,12 +73,25 @@ class PVector(object):
         return self
 
     def __len__(self):
+        """
+        >>> len(v(1, 2, 3))
+        3
+        """
         return self._count
 
     def __getitem__(self, index):
+        """
+        Get value at index. Full slicing support.
+
+        >>> v1 = v(5, 6, 7, 8)
+        >>> v1[2]
+        7
+        >>> v1[1:3]
+        pvector([6, 7])
+        """
         if isinstance(index, slice):
             # There are more conditions than the below where it would be OK to
-            # return ourself, implement those...
+            # return ourselves, implement those...
             if index.start is None and index.stop is None and index.step is None:
                 return self
 
@@ -87,15 +105,16 @@ class PVector(object):
         return self._node_for(index)[index & BIT_MASK]
 
     def __add__(self, other):
+        """
+        >>> v1 = v(1, 2)
+        >>> v2 = v(3, 4)
+        >>> v1 + v2
+        pvector([1, 2, 3, 4])
+        """
         return self.extend(other)
 
-    def __pvector_repr_with_recur_lock(self):
-        if len([f for f in extract_stack() if '__pvector_repr_with_recur_lock' in f[2]]) > 1:
-            return '(...)'
-        return str(self._totuple())
-
     def __repr__(self):
-        return self.__pvector_repr_with_recur_lock()
+        return 'pvector({})'.format(str(self._tolist()))
 
     __str__ = __repr__
 
@@ -129,6 +148,11 @@ class PVector(object):
         return self._tolist() <= other._tolist()
 
     def __mul__(self, times):
+        """
+        >>> v1 = v(1, 2)
+        >>> 3 * v1
+        pvector([1, 2, 1, 2, 1, 2])
+        """
         if times <= 0 or self is _EMPTY_VECTOR:
             return _EMPTY_VECTOR
         elif times == 1:
@@ -162,14 +186,31 @@ class PVector(object):
         return tuple(self._tolist())
 
     def __hash__(self):
+        """
+        >>> v1 = v(1, 2, 3)
+        >>> v2 = v(1, 2, 3)
+        >>> hash(v1) == hash(v2)
+        True
+        """
         # Taking the easy way out again...
         return hash(self._totuple())
 
-    def assoc(self, i, val):
+    def set(self, i, val):
         """
-        Return a new vector with element at position i replaced with val.
+        Return a new vector with element at position i replaced with val. The first vector remains unchanged.
+
+        Associng a value one step beyond the end of the vector is equal to appending. Associng beyond that will
+        result in an IndexError.
+
+        >>> v1 = v(1, 2, 3)
+        >>> v1.set(1, 4)
+        pvector([1, 4, 3])
+        >>> v1.set(3, 4)
+        pvector([1, 2, 3, 4])
+        >>> v1.set(-1, 4)
+        pvector([1, 2, 4])
         """
-        if not isinstance(i, (int, long)):
+        if not isinstance(i, Integral):
             raise TypeError("'%s' object cannot be interpreted as an index" % type(i).__name__)
 
         if i < 0:
@@ -181,20 +222,20 @@ class PVector(object):
                 new_tail[i & BIT_MASK] = val
                 return PVector(self._count, self._shift, self._root, new_tail)
 
-            return PVector(self._count, self._shift, self._do_assoc(self._shift, self._root, i, val), self._tail)
+            return PVector(self._count, self._shift, self._do_set(self._shift, self._root, i, val), self._tail)
 
         if i == self._count:
             return self.append(val)
 
         raise IndexError()
 
-    def _do_assoc(self, level, node, i, val):
+    def _do_set(self, level, node, i, val):
         ret = list(node)
         if level == 0:
             ret[i & BIT_MASK] = val
         else:
             sub_index = (i >> level) & BIT_MASK  # >>>
-            ret[sub_index] = self._do_assoc(level - SHIFT, node[sub_index], i, val)
+            ret[sub_index] = self._do_set(level - SHIFT, node[sub_index], i, val)
 
         return ret
 
@@ -226,6 +267,10 @@ class PVector(object):
     def append(self, val):
         """
         Return a new vector with val appended.
+
+        >>> v1 = v(1, 2)
+        >>> v1.append(3)
+        pvector([1, 2, 3])
         """
         if len(self._tail) < BRANCH_FACTOR:
             new_tail = list(self._tail)
@@ -267,7 +312,11 @@ class PVector(object):
     def extend(self, obj):
         """
         Return a new vector with all values in obj appended to it. Obj may be another
-        PVector or any other iterable.
+        PVector or any other Iterable.
+
+        >>> v1 = v(1, 2, 3)
+        >>> v1.extend([4, 5])
+        pvector([1, 2, 3, 4, 5])
         """
         # Mutates the new vector directly for efficiency but that's only an
         # implementation detail, once it is returned it should be considered immutable
@@ -302,20 +351,47 @@ class PVector(object):
         ret.append(self._new_path(level - SHIFT, tail_node))
         return ret
 
-    def assoc_in(self, keys, val):
+    def set_in(self, keys, val):
+        """
+        Insert val into nested persistent structure at position specified by Iterable keys. Any levels that
+        do not exist will be inserted as new PMaps.
+
+        >>> v1 = v(1, 2, m(a=5, b=6))
+        >>> v1.set_in((2, 'b'), 17)
+        pvector([1, 2, pmap({'a': 5, 'b': 17})])
+        >>> v1.set_in((2, 'c', 'd'), 17)
+        pvector([1, 2, pmap({'a': 5, 'c': pmap({'d': 17}), 'b': 6})])
+        """
         if not keys:
             return self
         elif len(keys) == 1:
-            return self.assoc(keys[0], val)
+            return self.set(keys[0], val)
         elif keys[0] == self._count:
-            return self.append(pmap().assoc_in(keys[1:], val))
+            return self.append(pmap().set_in(keys[1:], val))
         else:
-            return self.assoc(keys[0], self[keys[0]].assoc_in(keys[1:], val))
+            return self.set(keys[0], self[keys[0]].set_in(keys[1:], val))
 
     def index(self, value, *args, **kwargs):
+        """
+        Return first index of value. Additional indexes may be supplied to limit the search to a
+        sub range of the vector.
+        
+        >>> v1 = v(1, 2, 3, 4, 3)
+        >>> v1.index(3)
+        2
+        >>> v1.index(3, 3, 5)
+        4
+        """
         return self._tolist().index(value, *args, **kwargs)
 
     def count(self, value):
+        """
+        Return the number of times that value appears in the vector.
+
+        >>> v1 = v(1, 4, 3, 4)
+        >>> v1.count(4)
+        2
+        """
         return self._tolist().count(value)
 
 Sequence.register(PVector)
@@ -324,9 +400,13 @@ Hashable.register(PVector)
 _EMPTY_VECTOR = PVector(0, SHIFT, [], [])
 
 
-def _pvector(sequence=[]):
+def _pvector(sequence=()):
     """
-    Factory function, returns a new PVector object containing the elements in elements.
+    Factory function, returns a new PVector object containing the elements in sequence.
+
+    >>> v1 = pvector([1, 2, 3])
+    >>> v1
+    pvector([1, 2, 3])
     """
     return _EMPTY_VECTOR.extend(sequence)
 
@@ -335,7 +415,7 @@ pvector = _pvector
 try:
     # Use the C extension as underlying implementation if it is available
     from pvectorc import pvector as pvector_c
-    pvector = pvector_c
+    pvector = pvector
 except ImportError:
     pass
 
@@ -343,6 +423,10 @@ except ImportError:
 def v(*elements):
     """
     Factory function, returns a new PVector object containing all parameters.
+
+    >>> v1 = v(1, 2, 3)
+    >>> v1
+    pvector([1, 2, 3])
     """
     return pvector(elements)
 
@@ -350,7 +434,7 @@ def v(*elements):
 ####################### PMap #####################################
 class PMap(object):
     """
-    Do not instantiate directly, instead use the factory functions :py:func:`pmap` and :py:func:`pmapping` to
+    Do not instantiate directly, instead use the factory functions :py:func:`m` or :py:func:`pmap` to
     create an instance.
 
     Persistent map/dict. Tries to follow the same naming conventions as the built in dict where feasible.
@@ -361,18 +445,25 @@ class PMap(object):
     the containing vectors size the map is reallocated to a vector of double the size. This is done to avoid
     excessive hash collisions.
 
+    This structure corresponds most closely to the built in dict type and is intended as a replacement. Where the
+    semantics are the same (more or less) the same function names have been used but for some cases it is not possible,
+    for example assignments and deletion of values.
+
+    PMap implements the Mapping protocol and is Hashable.
+
     The following are examples of some common operations on persistent maps
 
-    >>> m = v(a=1, b=3)
-    >>> m2 = m.assoc('c', 3)
-    >>> m3 = m2.dissoc('a')
-    >>> m
-    {'a': 1, 'b': 3}
+    >>> m1 = m(a=1, b=3)
+    >>> m2 = m1.set('c', 3)
+    >>> m3 = m2.remove('a')
+    >>> m1
+    pmap({'a': 1, 'b': 3})
     >>> m2
-    {'a': 1, 'c': 3, 'b': 3}
+    pmap({'a': 1, 'c': 3, 'b': 3})
     >>> m3
-    {'c': 3, 'b': 3}
-    >>>
+    pmap({'c': 3, 'b': 3})
+    >>> m3['c']
+    3
     """
     __slots__ = ('_size', '_buckets')
 
@@ -443,18 +534,29 @@ class PMap(object):
         return self._size
 
     def __repr__(self):
-        return str(dict(self))
+        return 'pmap({})'.format(str(dict(self)))
 
     __eq__ = Mapping.__eq__
     __ne__ = Mapping.__ne__
     __str__ = __repr__
 
     def __hash__(self):
-        return hash(tuple(self.iteritems()))
+        # This hashing algorithm is probably not the speediest
+        return hash(frozenset(self.iteritems()))
 
-    def assoc(self, key, val):
+    def set(self, key, val):
         """
-        Return a new map with key and val inserted.
+        Return a new PMap with key and val inserted.
+
+        >>> m1 = m(a=1, b=2)
+        >>> m2 = m1.set('a', 3)
+        >>> m3 = m1.set('c' ,4)
+        >>> m1
+        pmap({'a': 1, 'b': 2})
+        >>> m2
+        pmap({'a': 3, 'b': 2})
+        >>> m3
+        pmap({'a': 1, 'c': 4, 'b': 2})
         """
         kv = (key, val)
         index, bucket = self._get_bucket(key)
@@ -465,23 +567,30 @@ class PMap(object):
                         return self
                     else:
                         new_bucket = [(k2, v2) if k2 != k else (k2, val) for k2, v2 in bucket]
-                        return PMap(self._size, self._buckets.assoc(index, new_bucket))
+                        return PMap(self._size, self._buckets.set(index, new_bucket))
 
             if len(self._buckets) < 0.67 * self._size:
-                return PMap(self._size, self._reallocate(2 * len(self._buckets))).assoc(key, val)
+                return PMap(self._size, self._reallocate(2 * len(self._buckets))).set(key, val)
             else:
                 new_bucket = [kv]
                 new_bucket.extend(bucket)
-                new_buckets = self._buckets.assoc(index, new_bucket)
+                new_buckets = self._buckets.set(index, new_bucket)
 
             return PMap(self._size + 1, new_buckets)
 
         # Skip reallocation check if there was no conflict
-        return PMap(self._size + 1, self._buckets.assoc(index, [kv]))
+        return PMap(self._size + 1, self._buckets.set(index, [kv]))
 
-    def dissoc(self, key):
+    def remove(self, key):
         """
-        Return a new map without the element specified by key.
+        Return a new PMap without the element specified by key. Returns reference to itself
+        if element is not present.
+
+        >>> m1 = m(a=1, b=2)
+        >>> m1.remove('a')
+        pmap({'b': 2})
+        >>> m1 is m1.remove('c')
+        True
         """
 
         # Should shrinking of the map ever be done if it becomes very small?
@@ -490,11 +599,19 @@ class PMap(object):
         if bucket:
             new_bucket = [(k, v) for (k, v) in bucket if k != key]
             if len(bucket) > len(new_bucket):
-                return PMap(self._size - 1, self._buckets.assoc(index, new_bucket if new_bucket else None))
+                return PMap(self._size - 1, self._buckets.set(index, new_bucket if new_bucket else None))
 
         return self
 
     def merge(self, *maps):
+        """
+        Return a new PMap with the items in Mappings inserted. If the same key is present in multiple
+        maps the rightmost (last) value is inserted.
+
+        >>> m1 = m(a=1, b=2)
+        >>> m1.merge(m(a=2, c=3), {'a': 17, 'd': 35})
+        pmap({'a': 17, 'c': 3, 'b': 2, 'd': 35})
+        """
         # Optimization opportunities here
         if not maps:
             return self
@@ -507,17 +624,25 @@ class PMap(object):
 
         result = self
         for k, v in merge_map.items():
-            result = result.assoc(k, v)
-        
+            result = result.set(k, v)
+
         return result
 
-    def assoc_in(self, keys, val):
+    def set_in(self, keys, val):
+        """
+        Insert val into nested persistent structure at position specified by Iterable keys. Any levels that
+        do not exist will be inserted as new PMaps.
+
+        >>> m1 = m(a=5, b=6, c=v(1, 2))
+        >>> m1.set_in(('c', 1), 17)
+        pmap({'a': 5, 'c': pvector([1, 17]), 'b': 6})
+        """
         if not keys:
             return self
         elif len(keys) == 1:
-            return self.assoc(keys[0], val)
+            return self.set(keys[0], val)
         else:
-            return self.assoc(keys[0], self.get(keys[0], _EMPTY_PMAP).assoc_in(keys[1:], val))
+            return self.set(keys[0], self.get(keys[0], _EMPTY_PMAP).set_in(keys[1:], val))
 
     def _reallocate_to_list(self, new_size):
         new_list = new_size * [None]
@@ -547,7 +672,7 @@ def _turbo_mapping(initial, pre_size):
         # key collisions
         initial = dict(initial)
 
-    for k, v in initial.iteritems():
+    for k, v in six.iteritems(initial):
         h = hash(k)
         index = h % size
         bucket = buckets[index]
@@ -570,6 +695,8 @@ def pmap(initial={}, pre_size=0):
     may have a positive performance impact in the cases where you know beforehand that a large number of elements
     will be inserted into the map eventually since it will reduce the number of reallocations required.
 
+    >>> pmap({'a': 13, 'b': '14'})
+    pmap({'a': 13, 'b': '14'})
     """
     if not initial:
         return _EMPTY_PMAP
@@ -580,29 +707,33 @@ def pmap(initial={}, pre_size=0):
 def m(**kwargs):
     """
     Factory function, inserts all key value arguments into the newly created map.
+
+    >>> m(a=13, b=14)
+    pmap({'a': 13, 'b': 14})
     """
     return pmap(kwargs)
 
-##################### Pset ########################
+##################### PSet ########################
 
 class PSet(object):
     """
-    Do not instantiate directly, instead use the factory function :py:func:`pset` to create an instance.
+    Do not instantiate directly, instead use the factory functions :py:func:`s` or :py:func:`pset`
+    to create an instance.
 
-    Persistent set implementation. Built on top of the persistent map.
+    Persistent set implementation. Built on top of the persistent map. The set supports all operations
+    in the Set protocol and is Hashable.
 
     Some examples:
 
     >>> s = pset([1, 2, 3, 1])
-    >>> s2 = s.assoc(4)
-    >>> s3 = s2.dissoc(2)
+    >>> s2 = s.add(4)
+    >>> s3 = s2.remove(2)
     >>> s
     pset([1, 2, 3])
     >>> s2
     pset([1, 2, 3, 4])
     >>> s3
     pset([1, 3, 4])
-    >>>
     """
     __slots__ = ('_map',)
 
@@ -619,7 +750,10 @@ class PSet(object):
         return len(self._map)
 
     def __repr__(self):
-        return 'p' + str(set(self))
+        if six.PY2 or not self:
+            return 'p' + str(set(self))
+
+        return 'pset([{}])'.format(str(set(self))[1:-1])
 
     __str__ = __repr__
 
@@ -631,11 +765,39 @@ class PSet(object):
         return PSet(pmap({k: True for k in it}, pre_size=pre_size))
 
     def add(self, element):
-        return PSet(self._map.assoc(element, True))
+        """
+        Return a new PSet with element added
 
-    def dissoc(self, element):
-        return PSet(self._map.dissoc(element))
+        >>> s1 = s(1, 2)
+        >>> s1.add(3)
+        pset([1, 2, 3])
+        """
+        return PSet(self._map.set(element, True))
 
+    def remove(self, element):
+        """
+        Return a new PSet with element removed. Raises KeyError if element is not present.
+
+        >>> s1 = s(1, 2)
+        >>> s1.remove(2)
+        pset([1])
+        """
+        if element in self._map:
+            return PSet(self._map.remove(element))
+
+        raise KeyError("Element '%s' not present in PSet" % element)
+
+    def discard(self, element):
+        """
+        Return a new PSet with element removed. Returns itself if element is not present.
+        """
+        if element in self._map:
+            return PSet(self._map.remove(element))
+
+        return self
+
+    # All the operations and comparisons you would expect on a set.
+    #
     # This is not very beautiful. If we avoid inheriting from PSet we can use the
     # __slots__ concepts (which requires a new style class) and hopefully save some memory.
     __le__ = Set.__le__
@@ -650,6 +812,13 @@ class PSet(object):
     __sub__ = Set.__sub__
     __xor__ = Set.__xor__
 
+    issubset = __le__
+    issuperset = __ge__
+    union = __or__
+    intersection = __and__
+    difference = __sub__
+    symmetric_difference = __xor__
+
     isdisjoint = Set.isdisjoint
 
 Set.register(PSet)
@@ -658,10 +827,14 @@ Hashable.register(PSet)
 _EMPTY_PSET = PSet(_EMPTY_PMAP)
 
 
-def pset(sequence=[], pre_size=8):
+def pset(sequence=(), pre_size=8):
     """
     Factory function, takes an iterable with elements to insert and optionally a sizing parameter equivalent to that
-    used for pmap().
+    used for :py:func:`pmap`.
+
+    >>> s1 = pset([1, 2, 3, 2])
+    >>> s1
+    pset([1, 2, 3])
     """
     if not sequence:
         return _EMPTY_PSET
@@ -670,4 +843,111 @@ def pset(sequence=[], pre_size=8):
 
 
 def s(*args):
+    """
+    Factory function for persistent sets.
+
+    >>> s1 = s(1, 2, 3, 2)
+    >>> s1
+    pset([1, 2, 3])
+    """
     return pset(args)
+
+
+######################################## Immutable object ##############################################
+
+
+def immutable(members='', name='Immutable', verbose=False):
+    """
+    Produces a class that either can be used standalone or as a base class for immutable classes.
+
+    This is a thin wrapper around a named tuple.
+
+    Constructing a type and using it to instantiate objects:
+
+    >>> Point = immutable('x, y', name='Point')
+    >>> p = Point(1, 2)
+    >>> p2 = p.set(x=3)
+    >>> p
+    Point(x=1, y=2)
+    >>> p2
+    Point(x=3, y=2)
+
+    Inheriting from a constructed type. In this case no type name needs to be supplied:
+
+    >>> class PositivePoint(immutable('x, y')):
+    ...     __slots__ = tuple()
+    ...     def __new__(cls, x, y):
+    ...         if x > 0 and y > 0:
+    ...             return super(PositivePoint, cls).__new__(cls, x, y)
+    ...         raise Exception('Coordinates must be positive!')
+    ...
+    >>> p = PositivePoint(1, 2)
+    >>> p.set(x=3)
+    PositivePoint(x=3, y=2)
+    >>> p.set(y=-3)
+    Traceback (most recent call last):
+    Exception: Coordinates must be positive!
+
+    The immutable class also supports the notion of frozen members. The value of a frozen members
+    cannot be updated. For example it could be used to implement an ID that should remain the same
+    over time. A frozen member is denoted by a trailing underscore.
+
+    >>> Point = immutable('x, y, id_', name='Point')
+    >>> p = Point(1, 2, id_=17)
+    >>> p.set(x=3)
+    Point(x=3, y=2, id_=17)
+    >>> p.set(id_=18)
+    Traceback (most recent call last):
+    AttributeError: Cannot set frozen members id_
+    """
+
+    if isinstance(members, six.string_types):
+        members = members.replace(',', ' ').split()
+
+    def frozen_member_test():
+        frozen_members = ["'%s'" % f for f in members if f.endswith('_')]
+        if frozen_members:
+            return """
+        frozen_fields = fields_to_modify & {{{frozen_members}}}
+        if frozen_fields:
+            raise AttributeError('Cannot set frozen members %s' % ', '.join(frozen_fields))
+            """.format(frozen_members=', '.join(frozen_members))
+
+        return ''
+
+    quoted_members = ', '.join("'%s'" % m for m in members)
+    template = """
+class {class_name}(namedtuple('ImmutableBase', [{quoted_members}], verbose={verbose})):
+    __slots__ = tuple()
+
+    def __repr__(self):
+        return super({class_name}, self).__repr__().replace('ImmutableBase', self.__class__.__name__)
+
+    def set(self, **kwargs):
+        if not kwargs:
+            return self
+
+        fields_to_modify = set(kwargs.keys())
+        if not fields_to_modify <= {member_set}:
+            raise AttributeError("'%s' is not a member" % ', '.join(fields_to_modify - {member_set}))
+
+        {frozen_member_test}
+
+        return self.__class__.__new__(self.__class__, *map(kwargs.pop, [{quoted_members}], self))
+    """.format(quoted_members=quoted_members,
+               member_set="{%s}" % quoted_members if quoted_members else 'set()',
+               frozen_member_test=frozen_member_test(),
+               verbose=verbose,
+               class_name=name)
+
+    if verbose:
+        print(template)
+
+    from collections import namedtuple
+    namespace = dict(namedtuple=namedtuple, __name__='pyrsistent_immutable')
+    try:
+        six.exec_(template, namespace)
+    except SyntaxError as e:
+        raise SyntaxError(e.message + ':\n' + template)
+
+    return namespace[name]
